/*
 *  Copyright 2005-2007 NVIDIA Corporation.  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of version 2 of the GNU General Public License as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it would be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write the Free Software Foundation, Inc., 59
 *  Temple Place - Suite 330, Boston MA 02111-1307, USA.
 */
#include "defs.h"
#include "tracing.h"
#include "errors.h"
#include "wn.h"
#include "bb.h"
#include "op.h"
#include "tn.h"
#include "cg.h"
#include "cgtarget.h"
#include <list>
#include <map>

static BOOL tracing = FALSE;
#define Trace(msg)	if (tracing) fprintf(TFile, msg "\n");

#if 0
#define COMPARE_32_OP(x)            (TOP_is_compare_32(OP_code(x)))
#else
#define COMPARE_32_OP(x)            (FALSE)
#endif
#define OUTPUT_CONVERTIBLE_16_OP(x) (TOP_is_convertible_to16_if_out16(OP_code(x)))
/* ignore compare operations till the bug is fixed in OCG */
#define INPUT_CONVERTIBLE_16_OP(x)  (COMPARE_32_OP(x) || \
			             TOP_is_convertible_to16_if_in16(OP_code(x)) )
#define LD_CONVERT16TO32_OP(x)      (TOP_is_load_16in32(OP_code(x)))
#define ST_CONVERT32TO16_OP(x)      (TOP_is_store_16from32(OP_code(x)))
#define CONVERT16TO32_OP(x)         (TOP_is_cvt_16to32(OP_code(x)))
#define CONVERT32TO16_OP(x)         (TOP_is_cvt_32to16(OP_code(x)))

#define CONVERTIBLE_16_OP(x)        (OUTPUT_CONVERTIBLE_16_OP(x) || \
                                     INPUT_CONVERTIBLE_16_OP(x) )
#define ALL_CONVERT16TO32_OP(x)     (LD_CONVERT16TO32_OP(x) ||  \
			                         CONVERT16TO32_OP(x) )
#define ALL_CONVERT32TO16_OP(x)     (ST_CONVERT32TO16_OP(x) ||  \
			                         CONVERT32TO16_OP(x) )

typedef struct TCO_TempInfo
{
	bool canBeDown;
	bool canBeUp;
	bool shouldNotBe;
	OP *tempOp; // We just need the result TN*
}TCO_TempInfo;

std::map<OP *, TCO_TempInfo *> TCO_TempInfoMap;
std::map<TOP, TOP> InstructionMap1;
std::map<TOP, TOP> InstructionMap2;

/******************************************************************/
/******************************************************************/

void Create_Instruction_Map(void)
{
  InstructionMap1[TOP_cvt_s16_s32] = TOP_mov_s16;
  InstructionMap1[TOP_cvt_u16_u32] = TOP_mov_u16;
  InstructionMap1[TOP_cvt_s32_s16] = TOP_mov_s16;
  InstructionMap1[TOP_cvt_u32_u16] = TOP_mov_u16;

  InstructionMap1[TOP_min_s32] = TOP_min_s16;
  InstructionMap1[TOP_max_s32] = TOP_max_s16;
  InstructionMap1[TOP_shr_s32_lit] = TOP_shr_s16_lit;
  //InstructionMap1[TOP_and_s32] = TOP_and_s16;
  //InstructionMap1[TOP_mov_s32] = TOP_mov_s16;
  //InstructionMap1[TOP_and_s32_lit] = TOP_and_s16_lit;
  //InstructionMap1[TOP_mov_s32_lit] = TOP_mov_s16_lit;
  InstructionMap1[TOP_min_u32] = TOP_min_u16;
  InstructionMap1[TOP_max_u32] = TOP_max_u16;
  InstructionMap1[TOP_shr_u32_lit] = TOP_shr_u16_lit;
  //InstructionMap1[TOP_and_u32] = TOP_and_u16;
  //InstructionMap1[TOP_mov_u32] = TOP_mov_u16;
  //InstructionMap1[TOP_and_u32_lit] = TOP_and_u16_lit;
  //InstructionMap1[TOP_mov_u32_lit] = TOP_mov_u16_lit;
  InstructionMap1[TOP_add_s32] = TOP_add_s16;
  InstructionMap1[TOP_sub_s32] = TOP_sub_s16;
  InstructionMap1[TOP_shl_s32_lit] = TOP_shl_s16_lit;
  InstructionMap1[TOP_and_s32] = TOP_and_s16;
  InstructionMap1[TOP_or_s32] = TOP_or_s16;
  InstructionMap1[TOP_xor_s32] = TOP_xor_s16;
  InstructionMap1[TOP_not_s32] = TOP_not_s16;
  InstructionMap1[TOP_mov_s32] = TOP_mov_s16;
  InstructionMap1[TOP_mul_lo_s32] = TOP_mul_lo_s16;
  InstructionMap1[TOP_add_s32_lit] = TOP_add_s16_lit;
  InstructionMap1[TOP_sub_s32_lit] = TOP_sub_s16_lit;
  InstructionMap1[TOP_and_s32_lit] = TOP_and_s16_lit;
  InstructionMap1[TOP_or_s32_lit] = TOP_or_s16_lit;
  InstructionMap1[TOP_xor_s32_lit] = TOP_xor_s16_lit;
  //InstructionMap1[TOP_not_s32_lit] = TOP_not_s16_lit;
  InstructionMap1[TOP_mov_s32_lit] = TOP_mov_s16_lit;
  InstructionMap1[TOP_mul_lo_s32_lit] = TOP_mul_lo_s16_lit;
  InstructionMap1[TOP_add_u32] = TOP_add_u16;
  InstructionMap1[TOP_sub_u32] = TOP_sub_u16;
  InstructionMap1[TOP_shl_u32_lit] = TOP_shl_u16_lit;
  InstructionMap1[TOP_and_u32] = TOP_and_u16;
  InstructionMap1[TOP_or_u32] = TOP_or_u16;
  InstructionMap1[TOP_xor_u32] = TOP_xor_u16;
  InstructionMap1[TOP_not_u32] = TOP_not_u16;
  InstructionMap1[TOP_mov_u32] = TOP_mov_u16;
  InstructionMap1[TOP_mul_lo_u32] = TOP_mul_lo_u16;
  InstructionMap1[TOP_add_u32_lit] = TOP_add_u16_lit;
  InstructionMap1[TOP_sub_u32_lit] = TOP_sub_u16_lit;
  InstructionMap1[TOP_and_u32_lit] = TOP_and_u16_lit;
  InstructionMap1[TOP_or_u32_lit] = TOP_or_u16_lit;
  InstructionMap1[TOP_xor_u32_lit] = TOP_xor_u16_lit;
  //InstructionMap1[TOP_not_u32_lit] = TOP_not_u16_lit;
  InstructionMap1[TOP_mov_u32_lit] = TOP_mov_u16_lit;
  InstructionMap1[TOP_mul_lo_u32_lit] = TOP_mul_lo_u16_lit;

  InstructionMap1[TOP_ld_global_u16_b32] = TOP_ld_global_u16;
  InstructionMap1[TOP_ld_global_s16_b32] = TOP_ld_global_s16;
  InstructionMap1[TOP_ld_global_u16_b32s] = TOP_ld_global_u16_s;
  InstructionMap1[TOP_ld_global_s16_b32s] = TOP_ld_global_s16_s;
  InstructionMap1[TOP_ld_global_u16_b32r] = TOP_ld_global_u16_r;
  InstructionMap1[TOP_ld_global_s16_b32r] = TOP_ld_global_s16_r;
  InstructionMap1[TOP_st_global_u16_b32] = TOP_st_global_u16;
  InstructionMap1[TOP_st_global_s16_b32] = TOP_st_global_s16;
  InstructionMap1[TOP_st_global_u16_b32s] = TOP_st_global_u16_s;
  InstructionMap1[TOP_st_global_s16_b32s] = TOP_st_global_s16_s;
  InstructionMap1[TOP_st_global_u16_b32r] = TOP_st_global_u16_r;
  InstructionMap1[TOP_st_global_s16_b32r] = TOP_st_global_s16_r;

  InstructionMap1[TOP_ld_shared_u16_b32] = TOP_ld_shared_u16;
  InstructionMap1[TOP_ld_shared_s16_b32] = TOP_ld_shared_s16;
  InstructionMap1[TOP_ld_shared_u16_b32s] = TOP_ld_shared_u16_s;
  InstructionMap1[TOP_ld_shared_s16_b32s] = TOP_ld_shared_s16_s;
  InstructionMap1[TOP_ld_shared_u16_b32r] = TOP_ld_shared_u16_r;
  InstructionMap1[TOP_ld_shared_s16_b32r] = TOP_ld_shared_s16_r;
  InstructionMap1[TOP_st_shared_u16_b32] = TOP_st_shared_u16;
  InstructionMap1[TOP_st_shared_s16_b32] = TOP_st_shared_s16;
  InstructionMap1[TOP_st_shared_u16_b32s] = TOP_st_shared_u16_s;
  InstructionMap1[TOP_st_shared_s16_b32s] = TOP_st_shared_s16_s;
  InstructionMap1[TOP_st_shared_u16_b32r] = TOP_st_shared_u16_r;
  InstructionMap1[TOP_st_shared_s16_b32r] = TOP_st_param_s16_r;

  InstructionMap1[TOP_ld_local_u16_b32] = TOP_ld_local_u16;
  InstructionMap1[TOP_ld_local_s16_b32] = TOP_ld_local_s16;
  InstructionMap1[TOP_ld_local_u16_b32s] = TOP_ld_local_u16_s;
  InstructionMap1[TOP_ld_local_s16_b32s] = TOP_ld_local_s16_s;
  InstructionMap1[TOP_ld_local_u16_b32r] = TOP_ld_local_u16_r;
  InstructionMap1[TOP_ld_local_s16_b32r] = TOP_ld_local_s16_r;
  InstructionMap1[TOP_st_local_u16_b32] = TOP_st_local_u16;
  InstructionMap1[TOP_st_local_s16_b32] = TOP_st_local_s16;
  InstructionMap1[TOP_st_local_u16_b32s] = TOP_st_local_u16_s;
  InstructionMap1[TOP_st_local_s16_b32s] = TOP_st_local_s16_s;
  InstructionMap1[TOP_st_local_u16_b32r] = TOP_st_local_u16_r;
  InstructionMap1[TOP_st_local_s16_b32r] = TOP_st_local_s16_r;

  InstructionMap1[TOP_ld_param_u16_b32] = TOP_ld_param_u16;
  InstructionMap1[TOP_ld_param_s16_b32] = TOP_ld_param_s16;
  InstructionMap1[TOP_ld_param_u16_b32s] = TOP_ld_param_u16_s;
  InstructionMap1[TOP_ld_param_s16_b32s] = TOP_ld_param_s16_s;
  InstructionMap1[TOP_ld_param_u16_b32r] = TOP_ld_param_u16_r;
  InstructionMap1[TOP_ld_param_s16_b32r] = TOP_ld_param_s16_r;
  InstructionMap1[TOP_st_param_u16_b32] = TOP_st_param_u16;
  InstructionMap1[TOP_st_param_s16_b32] = TOP_st_param_s16;
  InstructionMap1[TOP_st_param_u16_b32s] = TOP_st_param_u16_s;
  InstructionMap1[TOP_st_param_s16_b32s] = TOP_st_param_s16_s;
  InstructionMap1[TOP_st_param_u16_b32r] = TOP_st_param_u16_r;
  InstructionMap1[TOP_st_param_s16_b32r] = TOP_st_param_s16_r;

  InstructionMap1[TOP_ld_const_u16_b32] = TOP_ld_const_u16;
  InstructionMap1[TOP_ld_const_s16_b32] = TOP_ld_const_s16;
  InstructionMap1[TOP_ld_const_u16_b32s] = TOP_ld_const_u16_s;
  InstructionMap1[TOP_ld_const_s16_b32s] = TOP_ld_const_s16_s;
  InstructionMap1[TOP_ld_const_u16_b32r] = TOP_ld_const_u16_r;
  InstructionMap1[TOP_ld_const_s16_b32r] = TOP_ld_const_s16_r;

// Change to predicate version
#if 0
  InstructionMap1[TOP_set_eq_u32_s32_cc] = TOP_set_eq_u32_s16_cc;
  InstructionMap1[TOP_set_ne_u32_s32_cc] = TOP_set_ne_u32_s16_cc;
  InstructionMap1[TOP_set_lt_u32_s32_cc] = TOP_set_lt_u32_s16_cc;
  InstructionMap1[TOP_set_le_u32_s32_cc] = TOP_set_le_u32_s16_cc;
  InstructionMap1[TOP_set_gt_u32_s32_cc] = TOP_set_gt_u32_s16_cc;
  InstructionMap1[TOP_set_ge_u32_s32_cc] = TOP_set_ge_u32_s16_cc;
  InstructionMap1[TOP_set_eq_u32_u32_cc] = TOP_set_eq_u32_u16_cc;
  InstructionMap1[TOP_set_ne_u32_u32_cc] = TOP_set_ne_u32_u16_cc;
  InstructionMap1[TOP_set_lt_u32_u32_cc] = TOP_set_lt_u32_u16_cc;
  InstructionMap1[TOP_set_le_u32_u32_cc] = TOP_set_le_u32_u16_cc;
  InstructionMap1[TOP_set_gt_u32_u32_cc] = TOP_set_gt_u32_u16_cc;
  InstructionMap1[TOP_set_ge_u32_u32_cc] = TOP_set_ge_u32_u16_cc;
#endif

  InstructionMap2[TOP_mul_lo_s32] = TOP_mul24_lo_s32;
  InstructionMap2[TOP_mul_lo_u32] = TOP_mul24_lo_u32;
  InstructionMap2[TOP_mul_lo_s32_lit] = TOP_mul24_lo_s32_lit;
  InstructionMap2[TOP_mul_lo_u32_lit] = TOP_mul24_lo_u32_lit;
}

static void Clear_Memeory(void)
{
    InstructionMap1.clear();
    InstructionMap2.clear();

	std::map<OP *, TCO_TempInfo *>::iterator it;
	for(it=TCO_TempInfoMap.begin(); it!=TCO_TempInfoMap.end(); it++)
		free(it->second);
	TCO_TempInfoMap.clear();
}

static TCO_TempInfo *Create_TCO_TempInfo(void) 
{
  TCO_TempInfo *temp = NULL;
  temp = (TCO_TempInfo *) malloc( sizeof(TCO_TempInfo) );
  FmtAssert(temp != NULL, ("Can't allocate memory"));

  // Initially all are convertible ops
  temp->canBeDown = true; 
  temp->canBeUp = true; 
  temp->shouldNotBe = false;
  temp->tempOp = NULL;
  return temp;
}

static TCO_TempInfo *Get_TCO_TempInfo(OP *op) 
{
  std::map<OP *, TCO_TempInfo *>::iterator temp;
  if ( (temp = TCO_TempInfoMap.find(op)) == TCO_TempInfoMap.end() )
     return NULL;
  else
     return temp->second;
}

static void Set_TCO_TempInfo(OP *op, TCO_TempInfo *tinfo) 
{
  TCO_TempInfoMap.insert( pair<OP *, TCO_TempInfo *>(op, tinfo) );
}

// Reset the canBeDown flag in tempinfo
static void Reset_CanBeDown_Flag(OP *op)
{
  TCO_TempInfo *tempinfo;
  if( (tempinfo = Get_TCO_TempInfo(op)) == NULL )
    { Trace("Can't find the tempinfo for op in Reset_CanBeDown_Flag\n"); return; }
  tempinfo->canBeDown = false;  
}

// Reset the canBeUp flag in tempinfo
static void Reset_CanBeUp_Flag(OP *op)
{
  TCO_TempInfo *tempinfo;
  if( (tempinfo = Get_TCO_TempInfo(op)) == NULL )
    { Trace("Can't find the tempinfo for op in Reset_CanBeUp_Flag\n"); return; }
  tempinfo->canBeUp = false;  
}

// Set the shouldNotBe flag in tempinfo
// Set continue_pass to true if shouldNotBe is not set before
static void Set_ShouldNotBe_Flag(OP *op, bool *continue_pass)
{
  TCO_TempInfo *tempinfo;
  if( (tempinfo = Get_TCO_TempInfo(op)) == NULL )
    { Trace("Can't find the tempinfo for op in Set_ShouldNotBe_Flag\n"); return; }
  if( tempinfo->shouldNotBe == false )
	*continue_pass = true;
  tempinfo->shouldNotBe = true;  
}

// Return whether the operation can be converted
static bool Get_CanBe_Flag(OP *op)
{
  bool flag = false;
  TCO_TempInfo *tempinfo;
  if( (tempinfo = Get_TCO_TempInfo(op)) == NULL ) {
     Trace("Can't find the tempinfo for op in Get_CanBe_Flag\n");
	 return false; // Can't be converted
  }

  if( INPUT_CONVERTIBLE_16_OP(op) )
    flag |= tempinfo->canBeDown;
  if( OUTPUT_CONVERTIBLE_16_OP(op) ) 
    flag |= tempinfo->canBeUp;
  return flag;
}

// Return whether the operation can be and should be converted
static bool Get_CanBe_And_ShouldBe_Flag(OP *op)
{
  bool flag = false;
  TCO_TempInfo *tempinfo;
  if( (tempinfo = Get_TCO_TempInfo(op)) == NULL ) {
     Trace("Can't find the tempinfo for op in Get_CanBe_And_ShouldBe_Flag\n");
	 return false; // Can't be converted
  }

  if(tempinfo->shouldNotBe)
    return false;

  flag = Get_CanBe_Flag(op);
  return flag; //return flag && !tempinfo->shouldNotBe;
}

void Print_Op_TempInfo(OP *op)
{
  TCO_TempInfo *tempinfo = NULL;

  if(tracing) {
    Print_OP(op);
    if( (tempinfo = Get_TCO_TempInfo(op)) != NULL ) {
      if(INPUT_CONVERTIBLE_16_OP(op)) 
	    fprintf(TFile, "canBeDown: %d, ", tempinfo->canBeDown);
      if(OUTPUT_CONVERTIBLE_16_OP(op)) 
	    fprintf(TFile, "canBeUp: %d, ", tempinfo->canBeUp);
      fprintf(TFile, "shouldNotBe: %d ;", tempinfo->shouldNotBe);
      if(tempinfo->tempOp != NULL) Print_OP(tempinfo->tempOp);
	  else Trace("");
    }    
    else 
      Trace("");
  }
}

void Print_TempInfo()
{
  BB *bb = NULL;
  OP *op = NULL;
  TN *tn = NULL;
  
  Trace("\nPrinting temporary info associated with operations");
  Trace("**************************************************\n");
  for (bb = REGION_First_BB; bb != NULL; bb = BB_next(bb)) {
	 if (tracing) fprintf(TFile, "\n#########Print_TempInfo: bb %d##########\n", BB_id(bb));
     FOR_ALL_BB_OPs (bb, op) {
        Print_Op_TempInfo(op);
	 }
  }
  Trace("\n**************************************************\n");
}

// Check if a literal 4 byte integer constant can be 
// represented in lesser number of bytes
// sign = 0 mean unsigned demotion, target_size in number of bytes
static bool Check_For_Demotion(UINT32 src, UINT16 *target, bool sign, UINT8 target_size)
{
	bool flag;
    UINT32 mask1;
    UINT32 mask2, mask3;

	if(sign == 0) { // unsigned demotion
      mask1 = 0xffffffff;
	  mask1 = mask1 >> (32-target_size*8);
      if( mask1 == (src | mask1) ) {
	    flag = true; // 0x00XX pattern can be demoted 
	    *target = src;
	  }
	  else {
	  flag = false; // can't be demoted
	  }
	}
	else {          // signed demotion
      mask2 = 0xffffffff;
	  mask2 = mask2 >> (32-target_size*8+1);		  
	  mask3 = 0xffffffff;
      mask3 = mask3 << (32-target_size*8-1);
	  if( (mask2 == (src | mask2)) || (mask3 == (src & mask3)) ) {
        flag = true; // 0x000XX, 0xFF1XX pattern can be demoted
	    *target = src;
	  }
	  else {
	  flag = false; // can't be demoted
	  }
	}

return flag;
}

// convert cvt.u16_u32_32 into 2 explicit converts
// convert cvt.s16_s32_32 into 2 explicit converts
static void Convert_US16_US32_32(void)
{
  BB *bb = NULL;
  OP *op = NULL, *tmp_op1 = NULL, *tmp_op2 = NULL;
  TN *tn = NULL, *tmp_tn1 = NULL, *tmp_tn2 = NULL;
  
  Trace( "\nStarting Convert_US16_US32_32" );
  Trace( "*********************************\n" );

  for (bb = REGION_First_BB; bb != NULL; bb = BB_next(bb)) {
     if (tracing) fprintf(TFile, "\n#########convert_u16_u32_32: bb %d#########\n", BB_id(bb));
	 //Trace( "*********************************\n" );
     FOR_ALL_BB_OPs (bb, op) {

		 // Assuming, no condition code or predicate present in the operation
		 if( OP_code(op) == TOP_cvt_u16_u32_32 || OP_code(op) == TOP_cvt_s16_s32_32 ) {
           tmp_op1 =  Dup_OP(op);
		   tmp_tn1 = Build_RCLASS_TN(ISA_REGISTER_CLASS_integer16);
		   Set_TN_has_one_def(tmp_tn1);
		   Assign_Virtual_Register(tmp_tn1);
           // Assuming only 1 operand and 1 result
		   Set_OP_result(tmp_op1, 0, tmp_tn1);
		   BB_Insert_Op_Before(bb, op, tmp_op1);
		   Set_OP_opnd(op, 0, tmp_tn1);

		   if( OP_code(op) == TOP_cvt_u16_u32_32 ) {
		     OP_Change_Opcode(tmp_op1, TOP_cvt_u16_u32);
             OP_Change_Opcode(op, TOP_cvt_u32_u16);
		   }
		   else if( OP_code(op) == TOP_cvt_s16_s32_32 ) {
		     OP_Change_Opcode(tmp_op1, TOP_cvt_s16_s32);
             OP_Change_Opcode(op, TOP_cvt_s32_s16);
		   }
		 }
		 
	 }
  }

}

static void  TopDown_Set_CanBe_Pass(void)
{

  BB *bb = NULL;
  OP *op = NULL, *def_op = NULL;
  TN *tn = NULL;
  TCO_TempInfo *tempinfo = NULL;
  UINT16 target;
  UINT32 i;

  Trace( "\nStarting TopDown_Set_CanBe_Pass" );
  Trace( "*********************************\n" );

  for (bb = REGION_First_BB; bb != NULL; bb = BB_next(bb)) {
     if (tracing) fprintf(TFile, "\n#########TopDown_Set_CanBe_Pass: bb %d#########\n", BB_id(bb));
	 //Trace( "*********************************\n" );
     FOR_ALL_BB_OPs_FWD (bb, op) {
		if( INPUT_CONVERTIBLE_16_OP(op) ) {
		  // This operation can be converted to a 16-bit operation
          tempinfo = Create_TCO_TempInfo();
          Set_TCO_TempInfo(op, tempinfo); // Assuming no tempinfo already

          // Check all the inputs to of be 16-bit type 
          for (i = 0; i < OP_opnds(op); i++) {
	          tn = OP_opnd(op,i);
             if( TN_is_constant(tn) && !TN_has_value(tn) ) 
			   { Reset_CanBeDown_Flag(op); break; } // Not an integer constant
		     if( TN_is_constant(tn) && TN_has_value(tn) ) { // integer constant
			   if( !Check_For_Demotion(TN_value(tn), &target, \
				    TOP_is_convertible_to16_if_in16(OP_code(op)), 2) )
                 { Reset_CanBeDown_Flag(op); break; }
				 continue; 
			 }

			 // FTTB, handle only operands which are defined only once
			 // if( !TN_has_one_def(tn) ) { Reset_CanBeDown_Flag(op); break; }
		     
			 // search for unique reaching def			 
			 def_op = Find_Reaching_Def (tn, op);
			 // FTTB, handle only operands which have unique reaching def
			 if( def_op == NULL /*|| bb != OP_bb(def_op)*/ ) 
			   { Reset_CanBeDown_Flag(op); break; }
             
             // Reset canBeDown flag, if operand definition is not a 16to32 convert and not a convertible op itself
			 if( !ALL_CONVERT16TO32_OP(def_op) && !(CONVERTIBLE_16_OP(def_op) && Get_CanBe_And_ShouldBe_Flag(def_op)) )
               { Reset_CanBeDown_Flag(op); break; }

             // Till cvt.s16.s16 is supported, don't propagate to output
             if( COMPARE_32_OP(def_op) )
               { Reset_CanBeDown_Flag(op); break; }			 

#if 0
             // For input-only convertible operations, do a strict input type checking
             if( !OUTPUT_CONVERTIBLE_16_OP(op) ) 
			   // Here, op and def_op will be either input convertible or convert operations
               if( SIGNED_OP(op) != SIGNED_OP(def_op) ) {
                 Reset_CanBeDown_Flag(op); 
				 if(tracing) {
				   fprintf(TFile, "\nInput operand types don't match for input convertiple op:");
                   Print_OP(op);  Trace("");
				 }
			     break; 
			   }
#endif
		  }

		}          		
	 }
  }

}

static void  BottomUp_Set_CanBe_Pass(void)
{

  BB *bb = NULL;
  OP *op = NULL, *def_op = NULL;
  TN *tn = NULL;
  TCO_TempInfo *tempinfo = NULL;
  //UINT16 target;
  UINT32 i;

  Trace( "\nStarting BottomUp_Set_CanBe_Pass" );
  Trace( "*********************************\n" );

  for (bb = REGION_First_BB; bb != NULL; bb = BB_next(bb)) {
     if (tracing) fprintf(TFile, "\n#########BottomUp_Set_CanBe_Pass: bb %d#########\n", BB_id(bb));
	 //Trace( "*********************************\n" );
     FOR_ALL_BB_OPs_REV (bb, op) {

		 if( ALL_CONVERT32TO16_OP(op) || (CONVERTIBLE_16_OP(op) && Get_CanBe_And_ShouldBe_Flag(op)) ) {
          // output is/can be 16 bits
          // try to reset the canBeUp flags of the inputs 
          for (i = 0; i < OP_opnds(op); i++) {
	         tn = OP_opnd(op,i);

			 // Only consider the store value in store, ignore the store address
             if( ST_CONVERT32TO16_OP(op) && i != TOP_Find_Operand_Use(OP_code(op), OU_storeval) ) 
			   { continue; }               

			 if( TN_is_constant(tn) )  // literal constant
			   { continue; }

			 // FTTB, handle only operands which are defined only once
			 // if( !TN_has_one_def(tn) ) { Reset_CanBeUp_Flag(op); break; }
		     
			 // search for unique reaching def 
			 def_op = Find_Reaching_Def (tn, op);
			 // FTTB, handle only operands which have unique reaching def
			 // within the same basic block
			 if( def_op == NULL || bb != OP_bb(def_op) ) 
			   { continue; }

		     if( OUTPUT_CONVERTIBLE_16_OP(def_op) ) {
		       // This operation can be converted to a 16-bit operation
               if( (tempinfo = Get_TCO_TempInfo(def_op)) == NULL ) {
                 tempinfo = Create_TCO_TempInfo();
                 Set_TCO_TempInfo(def_op, tempinfo); 
			   }
			 }    
		  }

		}
		else {
          // output can't be 16 bits, reset the canBeUp flags of the inputs 
          for (i = 0; i < OP_opnds(op); i++) {
	         tn = OP_opnd(op,i);
			 if( TN_is_constant(tn) ) // literal constant
			   { continue; }
             // Change to predicate version
#if 0
             if( TN_register_class(tn) == ISA_REGISTER_CLASS_cc ) // Condtion code
		   	   { continue; }
#endif

			 // FTTB, handle only operands which are defined only once
			 // if( !TN_has_one_def(tn) ) { Reset_CanBeUp_Flag(op); break; }
		     
			 // search for unique reaching def 
			 def_op = Find_Reaching_Def (tn, op);
			 // FTTB, handle only operands which have unique reaching def
			 // within the same basic block
			 if( def_op == NULL || bb != OP_bb(def_op) ) 
			   { continue; }

		     if( CONVERTIBLE_16_OP(def_op) ) {
		       // This operation can't be converted to a 16-bit operation
               if( (tempinfo = Get_TCO_TempInfo(def_op)) == NULL ) {
                 tempinfo = Create_TCO_TempInfo();
                 Set_TCO_TempInfo(def_op, tempinfo); 
			   }
               Reset_CanBeUp_Flag(def_op);
			 }    
		  }

		}
			 
	 }
  }

}

static void Set_ShouldNotBe_Pass(void)
{

  BB *bb = NULL;
  OP *op = NULL, *def_op = NULL;
  TN *tn = NULL;
  //TCO_TempInfo *tempinfo;
  UINT16 target;
  UINT32 i;
  bool continue_pass, shouldBeFlag = true;

  Trace( "\nStarting Set_ShouldNotBe_Pass" );
  Trace( "*********************************\n" );

  for (bb = REGION_First_BB; bb != NULL; bb = BB_next(bb)) {
     if (tracing) fprintf(TFile, "\n#########Set_ShouldNotBe_Pass: bb %d#########\n", BB_id(bb));
	 //Trace( "*********************************\n" );

	 continue_pass = true;
	 while( continue_pass )
	 {
        continue_pass = false;

        FOR_ALL_BB_OPs_FWD (bb, op) {

           if( CONVERTIBLE_16_OP(op) && Get_CanBe_And_ShouldBe_Flag(op) ) {
             // Try to find out if ShouldNotBe flag will be set
             // Check all the inputs to of 16-bit type 
             for (i = 0; i < OP_opnds(op); i++) {
	            tn = OP_opnd(op,i);

                if( TN_is_constant(tn) && !TN_has_value(tn) ) // Not an integer constant
				  { Set_ShouldNotBe_Flag(op, &continue_pass); break; } 
		        if( TN_is_constant(tn) && TN_has_value(tn) ) { // integer constant
			      if( !Check_For_Demotion(TN_value(tn), &target, \
				          TOP_is_convertible_to16_if_in16(OP_code(op)), 2) )
				    { Set_ShouldNotBe_Flag(op, &continue_pass); break; }
				  continue; 
				}

			    // FTTB, handle only operands which are defined only once
			    // if( !TN_has_one_def(tn) ) { Reset_CanBeDown_Flag(op); break; }
		     
			    // search for unique reaching def			 
			    def_op = Find_Reaching_Def (tn, op);
			    // FTTB, handle only operands which have unique reaching def
			    if( def_op == NULL /*|| bb != OP_bb(def_op)*/ ) 
				  { Set_ShouldNotBe_Flag(op, &continue_pass); break; }
             
			    if( !ALL_CONVERT16TO32_OP(def_op) && !(CONVERTIBLE_16_OP(def_op) && Get_CanBe_And_ShouldBe_Flag(def_op)) ) 
                  { Set_ShouldNotBe_Flag(op, &continue_pass); break; } // This operation should not be converted

                // Till cvt.s16.s16 is supported, don't propagate to output
                if( COMPARE_32_OP(def_op) )
                  { Set_ShouldNotBe_Flag(op, &continue_pass); break; }		 

			 }
		   }
 
		   else if(  !ALL_CONVERT32TO16_OP(op) // Not a convert op
			      // Not a convertible op and its CanBeAndShouldBe Flag is not set
				  && !(CONVERTIBLE_16_OP(op) && Get_CanBe_And_ShouldBe_Flag(op)) ) {
		     //if( CONVERTIBLE_16_OP(op) && !Get_CanBe_And_ShouldBe_Flag(op) ) {
             //Propagate ShouldNotBeFlag to all inputs               
             for (i = 0; i < OP_opnds(op); i++) {
	            tn = OP_opnd(op,i);
			    if( TN_is_constant(tn) )  // literal constant
			      { continue; }
                // Change to predicate version
#if 0
			    if( TN_register_class(tn) == ISA_REGISTER_CLASS_cc ) // Condtion code
				  { continue; }
#endif
			
			    // FTTB, handle only operands which are defined only once
			    // if( !TN_has_one_def(tn) ) { Reset_CanBeUp_Flag(op); break; }
	     
                // search for unique reaching def 
		        def_op = Find_Reaching_Def (tn, op);
			    // FTTB, handle only operands which have unique reaching def
			    // within the same basic block
			    if( def_op == NULL || bb != OP_bb(def_op) ) 
				  { continue; }

                // Till cvt.s16.s16 is supported, don't propagate to input
                if( COMPARE_32_OP(def_op) )
                  { continue; }		 

		        if( CONVERTIBLE_16_OP(def_op) ) 
		          // This operation should not be converted to a 16-bit operation
                  { Set_ShouldNotBe_Flag(def_op, &continue_pass); }    
			 }
		   }           

		}
	 }
  }
	 
}

TN *Replace_Convert_16to32(OP *op)
{

  TN *tmp_tn = NULL;
  OP *tmp_op = NULL;
  TCO_TempInfo *tempinfo = NULL;

  if( (tempinfo = Get_TCO_TempInfo(op)) != NULL ) {
      if(tempinfo->tempOp == NULL)
	    { Trace("Null replacement operation in Replace_Convert_16to32. Possible error in earlier phase\n"); return NULL; }
	  else
		// Return the 16-bit operand of operation replacing convert
	    return OP_result(tempinfo->tempOp, 0); 
  }

  if( LD_CONVERT16TO32_OP(op) ) {
    tmp_op = Dup_OP(op);
	OP_Change_Opcode(tmp_op, InstructionMap1[OP_code(op)]);
    tmp_tn = Build_RCLASS_TN(ISA_REGISTER_CLASS_integer16);
	Set_TN_has_one_def(tmp_tn);
	Assign_Virtual_Register(tmp_tn);
    Set_OP_result(tmp_op, 0, tmp_tn);
	BB_Insert_Op_Before(OP_bb(op), op, tmp_op);	// replace by 16-bit load
  }
  else if( CONVERT16TO32_OP(op) ) {
    tmp_tn = OP_opnd(op, 0); // Return the 16-bit operand of convert
    tmp_op = Find_Reaching_Def (tmp_tn, op);
	if( !TN_has_one_def(tmp_tn) ) { // Ideally, we should check tmp_tn is not redefined
		                            // before, where it is going to be used after replacement
							        // Otherwise, add a copy to new temporary
	  // Add a move to a new temporary
      tmp_op = Dup_OP(op);
	  OP_Change_Opcode(tmp_op, InstructionMap1[OP_code(op)]);
      tmp_tn = Build_RCLASS_TN(ISA_REGISTER_CLASS_integer16);
	  Set_TN_has_one_def(tmp_tn);
	  Assign_Virtual_Register(tmp_tn);
      Set_OP_result(tmp_op, 0, tmp_tn);
	  BB_Insert_Op_Before(OP_bb(op), op, tmp_op);
	}
  }
  else {
    FmtAssert(FALSE, ("Unexpected operation in Replace_Convert_16to32. Error in earlier phase")); 
  }

  // Set the replacement operation in tempinfo
  tempinfo = Create_TCO_TempInfo();
  Set_TCO_TempInfo(op, tempinfo); // Assuming no tempinfo already
  tempinfo->tempOp = tmp_op;
	    
  if(tracing) fprintf(TFile, "\nReplaced this op. Temporary information    :");
  Print_Op_TempInfo(op);
  Trace("");

  return tmp_tn;
}

void Replace_Convert_32to16(OP *op)
{
  TN *tn = NULL;
  TN *tmp_tn = NULL;
  OP *tmp_op = NULL, *def_op = NULL, *rep_op = NULL;
  TCO_TempInfo *tempinfo = NULL, *tempinfo1 = NULL;

  if( ST_CONVERT32TO16_OP(op) ) // value stored is the second operand of store
    tn = OP_opnd(op, TOP_Find_Operand_Use(OP_code(op), OU_storeval) );
  else
    tn = OP_opnd(op, 0);

  if( TN_is_constant(tn) )  // literal constant
    { Trace("return position 1 in Replace_Convert_32to16"); return; } // Don't change the operation

  // FTTB, handle only operands which are defined only once
  // if( !TN_has_one_def(tn) ) { Reset_CanBeUp_Flag(op); break; }
		     
  // search for unique reaching def 
  def_op = Find_Reaching_Def (tn, op);
  // FTTB, handle only operands which have unique reaching def
  // within the same basic block
  if( def_op == NULL /*|| OP_bb(op) != OP_bb(def_op)*/ ) 
    { Trace("return position 2 in Replace_Convert_32to16"); return; } // Don't change the operation

  if( CONVERTIBLE_16_OP(def_op) && Get_CanBe_And_ShouldBe_Flag(def_op) ) {
    if( (tempinfo = Get_TCO_TempInfo(def_op)) == NULL ) {
      Trace("Null temporary information in Replace_Convert_32to16. Possible error in earlier phase\n"); return;
    }
	else { 
      if(tempinfo->tempOp == NULL)
	    { Trace("Null replacement operation in Replace_Convert_32to16. Possible error in earlier phase\n"); return; }
	  else { tmp_tn = OP_result(tempinfo->tempOp, 0); rep_op = tempinfo->tempOp; }
	}
  }
  else if( ALL_CONVERT16TO32_OP(def_op) ) { // Try to handle specially, if this is load. not to introduce any additional load
    tmp_tn = Replace_Convert_16to32(def_op); // tmp_tn = OP_opnd(def_op, 0); 
	if( tmp_tn == NULL ) {
      Trace("Null TN returned from Replace_Convert_16to32 in Replace_Convert_32to16. Possible error in earlier phase\n"); return;
	}
  }
  else
    { Trace("return position 3 in Replace_Convert_32to16"); return; } // Don't change the operation

  if( ST_CONVERT32TO16_OP(op) ) {
	OP_Change_Opcode(op, InstructionMap1[OP_code(op)]); // Replace by 16-bit store instruction
    Set_OP_opnd(op, TOP_Find_Operand_Use(OP_code(op), OU_storeval), tmp_tn);
  }
  else if( CONVERT32TO16_OP(op) ) {
    OP_Change_Opcode(op, InstructionMap1[OP_code(op)]); // Replace by move instruction
    Set_OP_opnd(op, 0/*TOP_Find_Operand_Use(OP_code(op), OU_opnd1)*/, tmp_tn);
  }
  else {
    FmtAssert(FALSE, ("Unexpected operation in Replace_Convert_32to16. Error in earlier phase")); 
  }

  if(tracing) fprintf(TFile, "\nReplaced this op. Temporary information    :");
  Print_Op_TempInfo(op);
  Trace("");
}

void Replace_Convertibe_Op(OP *op)
{

  UINT32 i;
  OP *def_op = NULL, *tmp_op = NULL;
  TN *tn = NULL, *tmp_tn[3] = {NULL}, *tmp_tn1 = NULL;
  TCO_TempInfo *tempinfo = NULL;
  UINT16 target;

  /* First create all the operands */
  for (i = 0; i < OP_opnds(op); i++) {
     tn = OP_opnd(op,i);
 
     if( TN_is_constant(tn) && !TN_has_value(tn) ) // Not an integer constant
	   { FmtAssert(FALSE, ("non-integer literal constant in Replace_Convertibe_Op. Error in earlier phase")); } 
	 if( TN_is_constant(tn) && TN_has_value(tn) ) { // integer constant
	   if( !Check_For_Demotion(TN_value(tn), &target, \
	       TOP_is_convertible_to16_if_in16(OP_code(op)), 2) )
	     { FmtAssert(FALSE, ("unconvertible integer constant in Replace_Convertibe_Op. Error in earlier phase")); }
       tmp_tn[i] = Gen_Unique_Literal_TN (target, 2);
	   continue;
	 }

	 // FTTB, handle only operands which are defined only once
	 // if( !TN_has_one_def(tn) ) { Reset_CanBeDown_Flag(op); break; }
		     
	 // search for unique reaching def			 
     def_op = Find_Reaching_Def (tn, op);
	 // FTTB, handle only operands which have unique reaching def
	 if( def_op == NULL /*|| OP_bb(op) != OP_bb(def_op)*/ ) 
	   { FmtAssert(FALSE, ("non-unique reaching def in Replace_Convertibe_Op. Error in earlier phase")); }
             
     if( ALL_CONVERT16TO32_OP(def_op) ) {
	   tmp_tn[i] = Replace_Convert_16to32(def_op); 
	   if( tmp_tn == NULL ) {
         Trace("Null TN returned from Replace_Convert_16to32 in Replace_Convertibe_Op. Possible error in earlier phase\n"); return;
	   }
	 }
	 else if ( CONVERTIBLE_16_OP(def_op) && Get_CanBe_And_ShouldBe_Flag(def_op) ) {   
       if( (tempinfo = Get_TCO_TempInfo(def_op)) == NULL ) {
	     Trace("Null temporary information in Replace_Convertibe_Op. Possible error in earlier phase\n"); return;
	   }
	   else { 
         if(tempinfo->tempOp == NULL)
	       { Trace("Null replacement operation in Replace_Convertibe_Op. Possible error in earlier phase\n"); return; }
	     else { tmp_tn[i] = OP_result(tempinfo->tempOp, 0); }
	   }
	 }
	 else { 
	   // don't assert as this shows up in slaswp test,
	   // but give up and continue.
	   // TODO:  real fix to avoid this situation
	   // FmtAssert(FALSE, ("Non-convertable operand in Replace_Convertibe_Op. Error in earlier phase")); 
	   DevWarn("Non-convertable operand in Replace_Convertibe_Op. Error in earlier phase, so giving up"); 
	   return;
	 }
  }

  // Next, create and insert the new instruction to replace
  if( COMPARE_32_OP(op) ) { 
    tmp_op = op; // For compare instructions like set, only change the operands
  }    
  else {
    tmp_op = Dup_OP(op);
	tmp_tn1 = Build_RCLASS_TN(ISA_REGISTER_CLASS_integer16);
	Set_TN_has_one_def(tmp_tn1);
    Assign_Virtual_Register(tmp_tn1);
    Set_OP_result(tmp_op, 0, tmp_tn1);
    BB_Insert_Op_Before(OP_bb(op), op, tmp_op);
  }
  for (i = 0; i < OP_opnds(op); i++) {
     Set_OP_opnd(tmp_op, i, tmp_tn[i]);
  }  
  OP_Change_Opcode(tmp_op, InstructionMap1[OP_code(op)]);

  // Set the replacement operation in tempinfo
  if( (tempinfo = Get_TCO_TempInfo(op)) == NULL ) {
	Trace("Null temporary information in Replace_Convertibe_Op. Possible error in earlier phase\n"); return;
  }
  else { 
    if(tempinfo->tempOp != NULL)
	  { Trace("Replacement operation already set in Replace_Convertibe_Op. Possible error in earlier phase\n"); return; }
	else { tempinfo->tempOp = tmp_op; }
  }
	    
  if(tracing) fprintf(TFile, "\nReplaced this op. Temporary information    :");
  Print_Op_TempInfo(op);
  Trace("");
}

void Generate_Mul_24(OP *op)
{  
  OP_Change_Opcode(op, InstructionMap2[OP_code(op)]);
  if(tracing) fprintf(TFile, "\nGenerating Mul-24 operation    :");
  Print_Op_TempInfo(op);
  Trace("");
}

void Remove_Convert_Ops(void) 
{

  BB *bb = NULL;
  OP *op, *def_op = NULL;
  TN *tn = NULL;
  TCO_TempInfo *tempinfo = NULL;
  //UINT16 target;
  UINT32 i;

  Trace( "\nStarting Remove_Convert_Ops" );
  Trace( "*********************************\n" );

  for (bb = REGION_First_BB; bb != NULL; bb = BB_next(bb)) {
     if (tracing) fprintf(TFile, "\n#########Remove_Convert_Ops: bb %d#########\n", BB_id(bb));
	 //Trace( "*********************************\n" );
     FOR_ALL_BB_OPs_FWD (bb, op) {
	    if(tracing) {
		  fprintf(TFile, "Considering op in Remove_Convert_Ops     :");
		  Print_OP(op);
		}
		if( CONVERTIBLE_16_OP(op) && Get_CanBe_And_ShouldBe_Flag(op) ) {
          Replace_Convertibe_Op(op);
 		}          	
		// 32-bit multiplcation whose CanBe value is set, but ShouldBe value is not set
		else if( TOP_is_mul_lo_32(OP_code(op)) && (!Get_CanBe_And_ShouldBe_Flag(op) && Get_CanBe_Flag(op)) ) {
		  Generate_Mul_24(op);
		}
		else if( ALL_CONVERT32TO16_OP(op) ) {
          Replace_Convert_32to16(op);
		}
	 }
  }
 
}

/**************************************************************************
Overview of the algorithm:
The algorithm maintains temporary information with appropriate operations.
It contains three flags. "CanBeDown" and "CanBeUp" flags indicates if the 
operation can be legally converted to 16-bit version. "ShouldNotBe" flag
indicates, the operation should not be converted to a 16-bit version as it 
may not be profitable. Temporary information also contains the 16-bit 
replacement operation. Operations, without temporary information, are 
considered illegal for conversion.

In the first step, The algorithm makes two passes over the basic block to 
determine the operations, which are legal to be converted to 16-bit version.
"Top-Down" pass sets the "CanBeDown" of input convertible operations, if all 
the inputs come from either convert operartions or input convertible operations
which have already been found eligible for conversion.
"Bottom-Up" pass sets the "CanBeUp" of output convertible operations, if its
output is fed to only convert operartions or input-output convertible operations,
which have already been found eligible for conversion.


Next, in another "Set_ShouldNotBe" pass, finds out which operations are not 
profitable for conversion. Here, we follow a simple rule for this. If all the 
inputs and outputs are either converts or elligible convertble operations, then the
operation is considered profitable, as it won't introduce any extra instructions.
Once, "ShouldNotBe" is set for an operation indicating it is ineligible for 
conversion, this information is propagated to make other operations ineligible
for conversion. This pass is repeated till no extra operation is found to be
ineligible.

In the last step, another pass is made through the operations. If it is an 
appropriate convert32to16(or store) operation, it is replaced with a move/store
instruction, as appropriate. If, it is an eligible convertible operation, it is 
replaced with an appropriate 16-bit operation. In both the cases, if the inputs 
come from appropriate convert16to32(or load) operation, depending upon the situation 
a load/move instruction is inserted or the its original 16-bit value is used 
as the operand in the new replaced operation of convert32to16/convertible 
operation. In any case, we maintain the new replaced operation in the temporary
information of each operation to avoid inserting duplicate operations.
**************************************************************************/
void Remove_Typeconv (void)
{
  BB *bb = NULL;
  OP *op = NULL;
  TN *tn = NULL;
  INT i;
  std::list<int> temp_oplist;

  tracing = Get_Trace(TP_TCO, 0x001);

  Trace("");
  /********************************************************/
  /************* Start of Type conversion code ****************/
  Trace( "Starting type conversion optimization\n" );
  Trace( "*************************************\n" );

  Create_Instruction_Map();

  Convert_US16_US32_32(); 

  TopDown_Set_CanBe_Pass();

  BottomUp_Set_CanBe_Pass();
  
  Set_ShouldNotBe_Pass();

  Print_TempInfo();

  Remove_Convert_Ops();

  Clear_Memeory();
  /************* End of Type conversion code ****************/
  /********************************************************/

}
