/*
 * Copyright 2005-2007 NVIDIA Corporation.  All rights reserved.
 */

/*

  This program is free software; you can redistribute it and/or modify it
  under the terms of version 2 of the GNU General Public License as
  published by the Free Software Foundation.

  This program is distributed in the hope that it would be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

  Further, this software is distributed without any warranty that it is
  free of the rightful claim of any third person regarding infringement 
  or the like.  Any license provided herein, whether implied or 
  otherwise, applies only to this software file.  Patent licenses, if 
  any, provided herein do not apply to combinations of this program with 
  other software, or any other product whatsoever.  

  You should have received a copy of the GNU General Public License along
  with this program; if not, write the Free Software Foundation, Inc., 59
  Temple Place - Suite 330, Boston MA 02111-1307, USA.

*/


//  
//  Generate ABI information
///////////////////////////////////////


//  $Revision: 1.22 $
//  $Date: 2001/03/10 01:15:59 $
//  $Author: mtibuild $
//  $Source: /osprey.src/osprey1.0/common/targ_info/abi/ia64/RCS/abi_properties.cxx,v $

#include <stddef.h>
#include "abi_properties_gen.h"
#include "targ_isa_registers.h"

static ABI_PROPERTY
	allocatable,
	unused,		// unused dummy result when only using cc
	tid,		// thread id
	ntid,		// number of thread ids
	ctaid,		// cta id
	nctaid,		// number of cta ids
	gridid,		// grid id
	clock,		// cycle counter
	pm,		// performance monitors
	physid;		// physical thread id


static const char *reg_names[ISA_REGISTER_MAX+1] = {
    "_",   NULL,   NULL,   NULL,   NULL,   NULL,   NULL,
  };

#define ISA_REGISTER_CLASS_Last_Reg(x) \
	ISA_REGISTER_CLASS_INFO_Last_Reg(ISA_REGISTER_CLASS_Info(x))

main()
{
  ABI_Properties_Begin("nvisa");

  allocatable = Create_Reg_Property("allocatable");
  unused = Create_Reg_Property("unused");
  tid = Create_Reg_Property("tid");
  ntid = Create_Reg_Property("ntid");
  ctaid = Create_Reg_Property("ctaid");
  nctaid = Create_Reg_Property("nctaid");
  gridid = Create_Reg_Property("gridid");
  clock = Create_Reg_Property("clock");
  pm = Create_Reg_Property("pm");
  physid = Create_Reg_Property("physid");

#if 0 // initially, nvisa has no calling convention
  callee = Create_Reg_Property("callee");
  caller = Create_Reg_Property("caller");
  func_arg = Create_Reg_Property("func_arg");
  func_val = Create_Reg_Property("func_val");
#endif 
#if 0 // initially, nvisa has no stack
  frame_ptr = Create_Reg_Property("frame_ptr");
  stack_ptr = Create_Reg_Property("stack_ptr");
#endif

  ///////////////////////////////////////
  Begin_ABI("nvisa");

  // Would be simpler to put these special regs at front,
  // but more user-friendly to put at end so user sees r1,r2,etc
  INT last_index = ISA_REGISTER_CLASS_Last_Reg(ISA_REGISTER_CLASS_integer16);
  INT tid_index = last_index - 2; // includes last index
  INT ntid_index = tid_index - 3;
  INT ctaid_index = ntid_index - 3;
  INT nctaid_index = ctaid_index - 3;
  INT gridid_index = nctaid_index - 1;
  reg_names[tid_index] = "%tid.x";
  reg_names[tid_index+1] = "%tid.y";
  reg_names[tid_index+2] = "%tid.z";
  reg_names[ntid_index] = "%ntid.x";
  reg_names[ntid_index+1] = "%ntid.y";
  reg_names[ntid_index+2] = "%ntid.z";
  reg_names[ctaid_index] = "%ctaid.x";
  reg_names[ctaid_index+1] = "%ctaid.y";
  reg_names[ctaid_index+2] = "%ctaid.z";
  reg_names[nctaid_index] = "%nctaid.x";
  reg_names[nctaid_index+1] = "%nctaid.y";
  reg_names[nctaid_index+2] = "%nctaid.z";
  reg_names[gridid_index] = "%gridid";
  Reg_Names(ISA_REGISTER_CLASS_integer16, 0, last_index, reg_names);

  last_index = ISA_REGISTER_CLASS_Last_Reg(ISA_REGISTER_CLASS_integer);
  INT clock_index = last_index;
  INT pm_index = clock_index - 4;
  INT physid_index = pm_index - 1;
  reg_names[clock_index] = "%clock";
  reg_names[pm_index] = "%pm0";
  reg_names[pm_index+1] = "%pm1";
  reg_names[pm_index+2] = "%pm2";
  reg_names[pm_index+3] = "%pm3";
  reg_names[physid_index] = "%physid";
  Reg_Names(ISA_REGISTER_CLASS_integer, 0, last_index, reg_names);

  Reg_Names(ISA_REGISTER_CLASS_integer64, 0, 1, reg_names);

  Reg_Property_Range (tid, ISA_REGISTER_CLASS_integer16, 
	tid_index, tid_index+2);
  Reg_Property_Range (ntid, ISA_REGISTER_CLASS_integer16, 
	ntid_index, ntid_index+2);
  Reg_Property_Range (ctaid, ISA_REGISTER_CLASS_integer16, 
	ctaid_index, ctaid_index+2);
  Reg_Property_Range (nctaid, ISA_REGISTER_CLASS_integer16, 
	nctaid_index, nctaid_index+2);
  Reg_Property (gridid, ISA_REGISTER_CLASS_integer16, gridid_index, -1);

  Reg_Property (clock, ISA_REGISTER_CLASS_integer, clock_index, -1);
  Reg_Property_Range (pm, ISA_REGISTER_CLASS_integer, 
	pm_index, pm_index+3);
  Reg_Property (physid, ISA_REGISTER_CLASS_integer, physid_index, -1);

  Reg_Property(unused, ISA_REGISTER_CLASS_integer, 0, -1);
  Reg_Property(unused, ISA_REGISTER_CLASS_integer16, 0, -1);
  Reg_Property(unused, ISA_REGISTER_CLASS_integer64, 0, -1);
  Reg_Property(unused, ISA_REGISTER_CLASS_float, 0, -1);
  Reg_Property(unused, ISA_REGISTER_CLASS_float64, 0, -1);


  // for now, no abi so don't need anything defined
  // except allocatable
  Reg_Property_Range (allocatable, ISA_REGISTER_CLASS_integer16, 1,
	gridid_index - 1); // ends before gridid

  Reg_Property_Range (allocatable, ISA_REGISTER_CLASS_integer, 1,
	physid_index - 1); // ends before physid

  Reg_Property_Range (allocatable, ISA_REGISTER_CLASS_integer64, 1,
	ISA_REGISTER_CLASS_Last_Reg(ISA_REGISTER_CLASS_integer64));

  Reg_Property_Range (allocatable, ISA_REGISTER_CLASS_float, 1,
	ISA_REGISTER_CLASS_Last_Reg(ISA_REGISTER_CLASS_float));

  Reg_Property_Range (allocatable, ISA_REGISTER_CLASS_float64, 1,
	ISA_REGISTER_CLASS_Last_Reg(ISA_REGISTER_CLASS_float64));

  Reg_Property_Range (allocatable, ISA_REGISTER_CLASS_predicate, 1, 
	ISA_REGISTER_CLASS_Last_Reg(ISA_REGISTER_CLASS_predicate));

  ABI_Properties_End();
}
